/**
 * SLPv2 user agent (UA) and service agent (SA) with GUI. Features:
 *
 *  (1) Configuration: TCP/UDP, fresh/incremental SrvReg
 *  (2) Five SLP messages: SrvRqst, SrvTypeRqst, AttrRqst, SrvReg, SrvDeReg
 *  (3) Three extensions: Mesh-Forwarding, Sort, Select
 *  (4) Status Windows
 *  (5) Available DA list: Auto DA discovery, select DA from the list
 *      (default is first DA), and remote DA discovery in given domains.
 *  (6) Failure handling: if no response from DA, try three times.
 * 
 * (c) Columbia University, 2001-2004, All Rights Reserved.
 * Author: Weibin Zhao
 */

import java.util.*;
import java.io.*;
import java.net.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.event.*;
import javax.naming.*;
import javax.naming.directory.*;

class ua extends JFrame implements ActionListener, ListSelectionListener {

    static int width = 620;
    static int height = 760;

    public static void main(String args[]) {
        try {
            // we need a trick to get the absolute domain name of the host
            String hname = (InetAddress.getLocalHost()).getHostAddress();
            hname = (InetAddress.getByName(hname)).getHostName();
            ua f = new ua(hname);
            f.setSize(width, height);
            f.show();
            if (args.length > 0) {
                f.append(args[0]);
            }
        } catch (Exception e) {
            System.err.println(e);
            System.exit(1);
        }
    }

    JMenuBar mbar;
    JMenu fileMenu, setupMenu, helpMenu;
    JMenuItem exitItem, aboutItem, batchReg; // simulate many SrvRegs
    String    batchSurfix = "";
    JRadioButtonMenuItem udpMode, tcpMode, freshReg, incReg;
                         // choose UPD/TCP mode, and fresh/incremental SrvReg 

    String[] mesgString = {"Service Request (SrvRqst)", 
                           "Service Type Request (StvTypeRqst)",
                           "Attribute Request (AttrRqst)",
                           "Service Registration (SrvReg)",
                           "Service Deregistration (SrvDeReg)"};
    String[] extString  = {"Sort",
                           "Select",
                           "Mesh Forwarding",
                           "Attribute List"};
    JComboBox  mesgList, extList;

    // APTN stands for Attribute, Predicate, Tag, or Naming Authority. 
    // As these four items can not appear in an SLPv2 message at the same
    // time, they share one GUI component.
    String[] aptnString = {" Predicate: ",
                           " Naming Authority: ",
                           " Tag List: ",
                           " Attribute: ",
                           " Tag List: "};
    JLabel     mesgLabel, srvtLabel, scopeLabel, ltagLabel, urlLabel,
               timeLabel, aptnLabel, daUrlLabel, daScopeLabel, dnsDomainLabel;
    JTextField srvtField, scopeField, ltagField, urlField, timeField, 
               aptnField, daUrlField, daScopeField, dnsDomainField;
    JTextArea  textarea;
    JList            daList;
    DefaultListModel daListModel;
    String SRVT, SCOPE, URL, LTAG, APTN, PR="", SPI="";
    int    LTIME, xid;
    MyTableModel myExtModel;
    JTable  table;
    JButton addExt, delExt, resetExt, clearFields, loadDefault,
            sendMesg, addDA, delDA, autoDA, remoteDA;
    Border compound;

    // Setup components
    uaFindDa       findDa;      // discover DA thread
    String         dirAgent;    // URL of current chosen DA
    String         tranMode = "udp";
    slpUdpHandler  udp;
    slpTcpHandler  tcp = null;
    int            slpFlag = Const.fresh_flag;

    // Basic operation classes
    slpMsgParser   parser;      // parser SLPv2 message
    slpMsgComposer composer;    // compose SLPv2 message
    uaAction       uac;         // UA actions for each SLPv2 message
    
    ua(String hname) {
        super("Welcome to mSLP UA/SA <"+hname+">");
        createGUI();

        // Setup UDP socket with the timeout for receive 
        try {
            DatagramSocket udpSocket = new DatagramSocket();
            udpSocket.setSoTimeout(Const.SoTimeout);
            udp = new slpUdpHandler(udpSocket);
        } catch (Exception e) {
            System.err.println(e);
        }

        // Starting DA discovery thread
        findDa = new uaFindDa(this);
        findDa.start();

        // Basic operation classes
        parser   = new slpMsgParser();
        composer = new slpMsgComposer();
        uac      = new uaAction(this);
    }

    public void append(String s) {
        textarea.append(s + "\n");
        int line = textarea.getLineCount();
        try {
            int pos = textarea.getLineStartOffset(line-1);
            textarea.setCaretPosition(pos);
        } catch (Exception ex) {
            System.out.println("Exception " + ex);
        }
    }

    public void actionPerformed(ActionEvent e) {
        if (e.getSource() == exitItem) {
            System.exit(0);
        } else if (e.getSource() == udpMode) {
            setUDPmode();
        } else if (e.getSource() == tcpMode) {
            setTCPmode();
        } else if (e.getSource() == freshReg) {
            setFreshReg();
        } else if (e.getSource() == incReg) {
            setIncReg();
        } else if (e.getSource() == sendMesg) {
            mesgHandler();
        } else if (e.getSource() == clearFields) {
            clearAllFields();
        } else if (e.getSource() == loadDefault) {
            loadDefaultValues();
        } else if (e.getSource() == batchReg) {
            batchRegSimulator();
        } else if (e.getSource() == mesgList) {
            aptnLabel.setText(aptnString[mesgList.getSelectedIndex()]);
            aptnField.setText("");
        } if (e.getSource() == addExt) {
            if (myExtModel.getRowCount() < myExtModel.getMaxRowCount()) {
                myExtModel.clearRow(myExtModel.getRowCount());
                myExtModel.setRowCount(myExtModel.getRowCount() + 1);
            }
        } else if (e.getSource() == delExt) {
            if (myExtModel.getRowCount() > 0) {
                if (table.getEditingRow() == (myExtModel.getRowCount() - 1)) {
                    table.removeEditor();
                }
                myExtModel.setRowCount(myExtModel.getRowCount() - 1);
            }
        } else if (e.getSource() == resetExt) {
            table.removeEditor();
            myExtModel.setRowCount(0);
        } else if (e.getSource() == addDA) {
            addDA(daUrlField.getText(), daScopeField.getText());
        } else if (e.getSource() == delDA) {
            delDA();
        } else if (e.getSource() == remoteDA) {
            remoteDA(dnsDomainField.getText());
        } else if (e.getSource() == autoDA) {
            findDa.discover(xid); // need to re-discover DA
            xid = (xid + 1) & 0xFFFF;
        }
    }    

    private void batchRegSimulator() {
            String s = urlField.getText();
            StringTokenizer st = new StringTokenizer(s, ":");
            if (st.countTokens() != 3) {
                append("Give batch URL as: 'dns-name:port#1:port#2'");
                return;
            }
            urlField.setText(st.nextToken()+":");
            int start = Integer.parseInt(st.nextToken());
            int end   = start + Integer.parseInt(st.nextToken());
            for (int i=start; i<end; i++) {
                batchSurfix = Integer.toString(i);
                mesgHandler();
            }
            urlField.setText(s);
            batchSurfix = "";
    }

/**
 * Setup transfer mode: UDP or TCP
 */
    private void setUDPmode() {
        tranMode = "udp";
        append("Switch to UDP transfer mode");
    }

    private void setTCPmode() {
        tranMode = "tcp";
        append("Switch to TCP transfer mode");
    }

/**
 * Setup to use fresh or incremental SrvReg
 */
    private void setFreshReg() {
        slpFlag = slpFlag | Const.fresh_flag;  // set fresh_flag
        append("Use fresh SrvReg");
    }

    private void setIncReg() {
        slpFlag = slpFlag & ~Const.fresh_flag; // clear fresh_flag
        append("Use incremental SrvReg");
    }

//---------------------------------------------------------------------
// Send an SLP message to the current DA (dirAgent), and process reply
//---------------------------------------------------------------------
    private void mesgHandler() {
        int id = mesgList.getSelectedIndex();
        if (dirAgent == null) { // choose to use DA, alternative to multicast
            append("Perform DA Discovery first, then choose one DA from the available DA list");
            return;
        }
        getNewPara();   // get all parameters from GUI
        byte[] mesgBuffer = null;
        switch (id) { // compose the sending message
            case 0: // SrvRqst
                mesgBuffer = composer.SrvRqst(xid, Const.normal_flag, LTAG, 
                                PR, SRVT, SCOPE, APTN, SPI);
                mesgBuffer = checkSSext(mesgBuffer);
                mesgBuffer = checkALext(mesgBuffer); // AttrList
                break;
            case 1: // SrvTypeRqst
                mesgBuffer = composer.SrvTypeRqst(xid, LTAG, PR, APTN, SCOPE);
                break;
            case 2: // AttrRqst
                String target = URL;    // try URL first, srvType second
                if (target.equals("")) target = SRVT;
                mesgBuffer = composer.AttrRqst(xid, LTAG, PR, target, SCOPE, 
                                                APTN, SPI);
                break;
            case 3: // SrvReg
                mesgBuffer = composer.SrvReg(xid, slpFlag & Const.fresh_flag,
                           LTAG, URL+batchSurfix, LTIME, SRVT, SCOPE, APTN);
                mesgBuffer = checkMFext(mesgBuffer);
                break;
            case 4: // SrvDeReg
                mesgBuffer = composer.SrvDeReg(xid, LTAG, SCOPE, URL, LTIME, 
                                             APTN);
                mesgBuffer = checkMFext(mesgBuffer);
                break;
        }       

        if (tranMode.equals("udp")) { // send message & process response
            uaUdpHandler(mesgBuffer);
        } else {
            uaTcpHandler(mesgBuffer);
        }
        xid = (xid + 1) & 0xFFFF;
    }

    private void getNewPara() {
        LTAG  = ltagField.getText();
        SCOPE = scopeField.getText();
        SRVT  = srvtField.getText();
        URL   = urlField.getText();
        LTIME = Integer.parseInt(timeField.getText());
        APTN   = aptnField.getText();
    }

    private void clearAllFields() {
        ltagField.setText("");
        scopeField.setText("");
        srvtField.setText("");
        urlField.setText("");
        timeField.setText("");
        aptnField.setText("");
    }

    private void loadDefaultValues() {
        ltagField.setText(Const.defaultLtag);
        scopeField.setText(Const.defaultScope);
        srvtField.setText(Const.defaultStype);
        urlField.setText(Const.defaultURL);
        timeField.setText(Const.defaultLtime);
        aptnField.setText("");
    }

    private byte[] checkMFext(byte[] mesgBuffer) { // Mesh Forwarding extension 
        int n = myExtModel.getRowCount();
        for (int i=0; i<n; i++) {
            String ext_name = (String) myExtModel.getValueAt(i, 1);
            if (ext_name.equals(extString[2])) {
                return composer.MeshFwdExt(mesgBuffer, Const.RqstFwd,
                                System.currentTimeMillis(), dirAgent, 0);
            }
        }
        return mesgBuffer;
    }

    private byte[] checkALext(byte[] mesgBuffer) { // Attribute List extension 
        int n = myExtModel.getRowCount();
        for (int i=0; i<n; i++) {
            String ext_name = (String) myExtModel.getValueAt(i, 1);
            if (ext_name.equals(extString[3])) {
                return composer.AttrListExt(mesgBuffer, null);
            }
        }
        return mesgBuffer;
    }

    private byte[] checkSSext(byte[] mesgBuffer) { // Select & Sort
        int n = myExtModel.getRowCount();
        byte[] buf = mesgBuffer;
        for (int i=0; i<n; i++) {
            String ext_name = (String) myExtModel.getValueAt(i, 1);
            String ext_val  = (String) myExtModel.getValueAt(i, 2);
            if (ext_name.equals(extString[0])) {        // Sort
                String key = ext_val.trim();
                buf = composer.SortExt(buf, key);
            } else if (ext_name.equals(extString[1])) { // Select
                int bound = (new Integer(ext_val.trim())).intValue();
                buf = composer.SelectExt(buf, bound);
            }
        }
        return buf;
    }

/**
 * Try to send a TCP message to DA, and process the response from DA
 * If no response, then give up
 */
    private void uaTcpHandler(byte[] mesgBuffer) {
        if (mesgBuffer == null) return;
        if (tcp == null) connectDA();
        if (tcp == null) return;
        tcp.send(mesgBuffer, mesgBuffer.length);
        byte[] recvMesg;
        if ((recvMesg = tcp.getHeader()) == null) { // TimeOut
            daFail();
            return;
        }
        parser.Header(recvMesg);
        if ((recvMesg = tcp.getBody(parser.getPacketLen())) == null) { //TimeOut
            daFail();
            return;
        }
        uac.action(parser, recvMesg, 0);
    }

/**
 * Try to send a UDP message to DA, and process the response from DA
 * If no response, try at most three times, then give up
 */
    private void uaUdpHandler(byte[] mesgBuffer) {
        if (mesgBuffer == null) return;
        for (int i=1; i<=Const.maxTryNum; i++) {
            udp.send(mesgBuffer, Util.url2dname(dirAgent), Const.port);
            byte[] recvMesg = udp.receive();
            if (recvMesg == null) continue;
            parser.Header(recvMesg);
            uac.action(parser, recvMesg, Const.header_len);
            return;
        }
        daFail();
    }

/**
 * make a TCP connection for TCP transfer mode
 */
    private void connectDA() {
        try {
            InetAddress addr = InetAddress.getByName(Util.url2dname(dirAgent));
            append("Connect to: "+addr.toString()+":"+Const.port);
            Socket socket = new Socket(addr, Const.port);
            socket.setSoTimeout(Const.SoTimeout);
            tcp      = new slpTcpHandler(socket, null);
        } catch (ConnectException e) {
            System.err.println(e.getMessage());
        } catch (Exception e) {
            System.err.println(e);
        }
    }

//----------------------
// Create GUI for UA/SA
//----------------------
    private void createGUI() {

        fileMenu = new JMenu("File");
        exitItem = new JMenuItem("Exit");
        exitItem.addActionListener(this);
        fileMenu.add(exitItem);

// For simulation
//      batchReg = new JMenuItem("Bulk SrvReg Simulator");
//      batchReg.addActionListener(this);
//      fileMenu.add(batchReg);

        setupMenu = new JMenu("Configuration");

        ButtonGroup transGroup = new ButtonGroup();
        udpMode = new JRadioButtonMenuItem("Use UDP");
        udpMode.setSelected(true);
        udpMode.addActionListener(this);
        transGroup.add(udpMode);
        setupMenu.add(udpMode);

        tcpMode = new JRadioButtonMenuItem("Use TCP");
        tcpMode.addActionListener(this);
        transGroup.add(tcpMode);
        setupMenu.add(tcpMode);

        setupMenu.addSeparator();
        ButtonGroup regGroup = new ButtonGroup();
        freshReg = new JRadioButtonMenuItem("Use Fresh SrvReg");
        freshReg.setSelected(true);
        freshReg.addActionListener(this);
        regGroup.add(freshReg);
        setupMenu.add(freshReg);

        incReg = new JRadioButtonMenuItem("Use Incremental SrvReg");
        incReg.addActionListener(this);
        regGroup.add(incReg);
        setupMenu.add(incReg);

        helpMenu = new JMenu("Help");

        aboutItem = new JMenuItem("About...");
        aboutItem.addActionListener(this);
        helpMenu.add(aboutItem);

        mbar = new JMenuBar();
        Border raisedbevel, loweredbevel;
        raisedbevel = BorderFactory.createRaisedBevelBorder();
        loweredbevel = BorderFactory.createLoweredBevelBorder();
        compound = BorderFactory.createCompoundBorder(
                                raisedbevel, loweredbevel);
        mbar.setBorder(compound);
        mbar.add(fileMenu);
        mbar.add(setupMenu);
        mbar.add(helpMenu);
//      mbar.setHelpMenu(helpMenu); // Not implemented by Java Swing
        setJMenuBar(mbar);

        GridBagLayout gridbag = new GridBagLayout();
        getContentPane().setLayout(gridbag);
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.weightx = 1.0;
        c.weighty = 0.0;
        c.gridwidth = 1;
        c.gridheight = 1;

        c.gridx = 0;
        c.gridy = 0;
        JScrollPane obj = Message_Composer();
        gridbag.setConstraints(obj, c);
        getContentPane().add(obj);

        c.gridy = 1;
        obj = Extension_Composer();
        gridbag.setConstraints(obj, c);
        getContentPane().add(obj);

        c.weighty = 1.0;
        c.gridy = 2;
        obj = Status_Window();
        gridbag.setConstraints(obj, c);
        getContentPane().add(obj);

        c.weighty = 0.0;
        c.gridy = 3;
        obj = DA_List();
        gridbag.setConstraints(obj, c);
        getContentPane().add(obj);
    }

    private JScrollPane Message_Composer() {

        JPanel mesgPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        mesgPanel.setLayout(gridbag);
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.weighty = 0.0;

        c.weightx = 0.33;
        mesgLabel = new JLabel(" Message Type: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 0;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(mesgLabel, c);
        mesgPanel.add(mesgLabel);

        c.weightx = 0.66;
        mesgList = new JComboBox(mesgString);
        mesgList.setSelectedIndex(0);
        mesgList.addActionListener(this);
        c.gridx = 1;
        c.gridy = 0;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(mesgList, c);
        mesgPanel.add(mesgList);

        c.weightx = 0.33;
        scopeLabel = new JLabel(" Scope: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 1;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(scopeLabel, c);
        mesgPanel.add(scopeLabel);

        c.weightx = 0.66;
        scopeField = new JTextField(Const.defaultScope);
        scopeField.setForeground(Color.black);
        scopeField.setBackground(Color.white);
        scopeField.addActionListener(this);
        c.gridx = 1;
        c.gridy = 1;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(scopeField, c);
        mesgPanel.add(scopeField);

        c.weightx = 0.33;
        srvtLabel = new JLabel(" Service Type: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 2;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(srvtLabel, c);
        mesgPanel.add(srvtLabel);

        c.weightx = 0.66;
        srvtField = new JTextField(Const.defaultStype);
        srvtField.setForeground(Color.black);
        srvtField.setBackground(Color.white);
        srvtField.addActionListener(this);
        c.gridx = 1;
        c.gridy = 2;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(srvtField, c);
        mesgPanel.add(srvtField);

        c.weightx = 0.33;
        urlLabel = new JLabel(" URL: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 3;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(urlLabel, c);
        mesgPanel.add(urlLabel);

        c.weightx = 0.66;
        urlField = new JTextField(Const.defaultURL);
        urlField.setForeground(Color.black);
        urlField.setBackground(Color.white);
        urlField.addActionListener(this);
        c.gridx = 1;
        c.gridy = 3;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(urlField, c);
        mesgPanel.add(urlField);

        c.weightx = 0.33;
        timeLabel = new JLabel(" Life Time: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 4;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(timeLabel, c);
        mesgPanel.add(timeLabel);

        c.weightx = 0.66;
        timeField = new JTextField(Const.defaultLtime);
        timeField.setForeground(Color.black);
        timeField.setBackground(Color.white);
        timeField.addActionListener(this);
        c.gridx = 1;
        c.gridy = 4;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(timeField, c);
        mesgPanel.add(timeField);

        c.weightx = 0.33;
        ltagLabel = new JLabel(" Language Tag: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 5;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(ltagLabel, c);
        mesgPanel.add(ltagLabel);

        c.weightx = 0.66;
        ltagField = new JTextField(Const.defaultLtag);
        ltagField.setForeground(Color.black);
        ltagField.setBackground(Color.white);
        ltagField.addActionListener(this);
        c.gridx = 1;
        c.gridy = 5;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(ltagField, c);
        mesgPanel.add(ltagField);

        c.weightx = 0.33;
        aptnLabel = new JLabel(" Predicate: ", JLabel.CENTER);
        c.gridx = 0;
        c.gridy = 6;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(aptnLabel, c);
        mesgPanel.add(aptnLabel);

        c.weightx = 0.66;
        aptnField = new JTextField("");
        aptnField.setForeground(Color.black);
        aptnField.setBackground(Color.white);
        aptnField.addActionListener(this);
        c.gridx = 1;
        c.gridy = 6;
        c.gridwidth = 2;
        c.gridheight = 1;
        gridbag.setConstraints(aptnField, c);
        mesgPanel.add(aptnField);

        clearFields    = new JButton("Clear All Fields");
        clearFields.addActionListener(this);
        loadDefault  = new JButton("Load Default Values");
        loadDefault.addActionListener(this);
        sendMesg = new JButton("Send Message");
        sendMesg.addActionListener(this);

        c.weightx = 0.33;
        c.gridx = 0;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(clearFields, c);
        mesgPanel.add(clearFields);

        c.weightx = 0.33;
        c.gridx = 1;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(loadDefault, c);
        mesgPanel.add(loadDefault);

        c.weightx = 0.33;
        c.gridx = 2;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(sendMesg, c);
        mesgPanel.add(sendMesg);

        JScrollPane MesgPanel = new JScrollPane(mesgPanel);
        MesgPanel.setBorder(compound);
        JLabel mesgLabel = new JLabel("Message Composer", JLabel.CENTER);
        mesgLabel.setForeground(Color.blue);
        mesgLabel.setBackground(Color.cyan);
        mesgLabel.setOpaque(true);
        MesgPanel.setColumnHeaderView(mesgLabel);
        MesgPanel.setPreferredSize(new Dimension(600, 205));
        return MesgPanel;
    }

    private JScrollPane Extension_Composer() {

        myExtModel = new MyTableModel();
        table = new JTable(myExtModel);
        table.setPreferredScrollableViewportSize(new Dimension(570, 65));
        table.getColumnModel().getColumn(0).setPreferredWidth(130);
        table.getColumnModel().getColumn(1).setPreferredWidth(120);
        table.getColumnModel().getColumn(2).setPreferredWidth(320);
        TableColumn ext_name = table.getColumnModel().getColumn(1);
        extList = new JComboBox(extString);
        extList.setSelectedIndex(1);
        ext_name.setCellEditor(new DefaultCellEditor(extList));
        table.setDefaultRenderer(Integer.class, new LabelRender());
        JScrollPane tablePanel = new JScrollPane(table);

        addExt    = new JButton("Append New Extension");
        addExt.addActionListener(this);
        delExt = new JButton("Remove Last Extension");
        delExt.addActionListener(this);
        resetExt  = new JButton("Remove All Extensions");
        resetExt.addActionListener(this);

        JPanel extensionPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        extensionPanel.setLayout(gridbag);
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.weighty = 0.0;

        c.weightx = 1.0;
        c.gridx = 0;
        c.gridy = 0;
        c.gridwidth = 3;
        c.gridheight = 4;
        gridbag.setConstraints(tablePanel, c);
        extensionPanel.add(tablePanel);

        c.weightx = 0.33;
        c.gridx = 0;
        c.gridy = 4;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(addExt, c);
        extensionPanel.add(addExt);

        c.gridx = 1;
        c.gridy = 4;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(delExt, c);
        extensionPanel.add(delExt);

        c.gridx = 2;
        c.gridy = 4;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(resetExt, c);
        extensionPanel.add(resetExt);

        JScrollPane ExtensionPanel = new JScrollPane(extensionPanel);
        ExtensionPanel.setBorder(compound);
        JLabel extensionLabel = new JLabel("Extension Composer", JLabel.CENTER);
        extensionLabel.setForeground(Color.blue);
        extensionLabel.setBackground(Color.cyan);
        extensionLabel.setOpaque(true);
        ExtensionPanel.setColumnHeaderView(extensionLabel);
        ExtensionPanel.setPreferredSize(new Dimension(600, 140));
        return ExtensionPanel;
    }

    class LabelRender extends JLabel implements TableCellRenderer {

        LabelRender() {
            super();
            setOpaque(true);
        }

        public Component getTableCellRendererComponent(
                                JTable table, Object val, 
                                boolean isSelected, boolean hasFocus,
                                int row, int column) {
            setText("<"+val.toString()+">");
            setBackground(Color.lightGray);
            setHorizontalAlignment(SwingConstants.CENTER);
            return this;
        }
    }

    private JScrollPane Status_Window() {
        textarea = new JTextArea();
        textarea.setEditable(false);
        textarea.setForeground(Color.black);
        textarea.setBackground(Color.lightGray);
        JScrollPane textPanel = new JScrollPane(textarea);
        textPanel.setBorder(compound);
        JLabel statusLabel = new JLabel("Status Window", JLabel.CENTER);
        statusLabel.setForeground(Color.blue);
        statusLabel.setBackground(Color.cyan);
        statusLabel.setOpaque(true);
        textPanel.setColumnHeaderView(statusLabel);
        textPanel.setPreferredSize(new Dimension(600, 100));
        return textPanel;
    }

    private JScrollPane DA_List() {
        daListModel = new DefaultListModel();
        daList = new JList(daListModel);
        daList.setBackground(Color.lightGray);
        daList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        daList.addListSelectionListener(this);
        JScrollPane listPanel = new JScrollPane(daList);
        listPanel.setPreferredSize(new Dimension(600, 73));

        addDA = new JButton("Add New DA");
        addDA.addActionListener(this);
        delDA = new JButton("Remove Selected DA");
        delDA.addActionListener(this);
        autoDA = new JButton("Auto DA Discovery");
        autoDA.addActionListener(this);
        remoteDA = new JButton("Remote DA Discovery");
        remoteDA.addActionListener(this);
        daUrlLabel = new JLabel(" New DA URL: ", JLabel.CENTER);
        daUrlField = new JTextField();
        daScopeLabel = new JLabel(" New DA Scope: ", JLabel.CENTER);
        daScopeField = new JTextField();
        dnsDomainLabel = new JLabel(" Remote DNS Domain: ", JLabel.CENTER);
        dnsDomainField = new JTextField();

        JPanel daPanel = new JPanel();
        GridBagLayout gridbag = new GridBagLayout();
        daPanel.setLayout(gridbag);
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.weighty = 0.0;

        c.weightx = 1.0;
        c.weighty = 1.0;
        c.gridx = 0;
        c.gridy = 0;
        c.gridwidth = 4;
        c.gridheight = 4;
        gridbag.setConstraints(listPanel, c);
        daPanel.add(listPanel);

        c.weighty = 0.0;

        c.weightx = 0.25;
        c.gridx = 0;
        c.gridy = 4;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(daUrlLabel, c);
        daPanel.add(daUrlLabel);

        c.weightx = 0.75;
        c.gridx = 1;
        c.gridy = 4;
        c.gridwidth = 3;
        c.gridheight = 1;
        gridbag.setConstraints(daUrlField, c);
        daPanel.add(daUrlField);

        c.weightx = 0.25;
        c.gridx = 0;
        c.gridy = 5;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(daScopeLabel, c);
        daPanel.add(daScopeLabel);

        c.weightx = 0.75;
        c.gridx = 1;
        c.gridy = 5;
        c.gridwidth = 3;
        c.gridheight = 1;
        gridbag.setConstraints(daScopeField, c);
        daPanel.add(daScopeField);

        c.weightx = 0.25;
        c.gridx = 0;
        c.gridy = 6;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(dnsDomainLabel, c);
        daPanel.add(dnsDomainLabel);

        c.weightx = 0.75;
        c.gridx = 1;
        c.gridy = 6;
        c.gridwidth = 3;
        c.gridheight = 1;
        gridbag.setConstraints(dnsDomainField, c);
        daPanel.add(dnsDomainField);

        c.weightx = 0.25;
        c.gridx = 0;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(addDA, c);
        daPanel.add(addDA);

        c.gridx = 1;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(delDA, c);
        daPanel.add(delDA);

        c.gridx = 2;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(remoteDA, c);
        daPanel.add(remoteDA);

        c.gridx = 3;
        c.gridy = 7;
        c.gridwidth = 1;
        c.gridheight = 1;
        gridbag.setConstraints(autoDA, c);
        daPanel.add(autoDA);

        JScrollPane ListPanel = new JScrollPane(daPanel);
        ListPanel.setBorder(compound);
        JLabel dalistLabel = new JLabel("Available DA List", JLabel.CENTER);
        dalistLabel.setForeground(Color.blue);
        dalistLabel.setBackground(Color.cyan);
        dalistLabel.setOpaque(true);
        ListPanel.setColumnHeaderView(dalistLabel);
        ListPanel.setPreferredSize(new Dimension(600, 213));
        return ListPanel;
    }

/**
 * Available DA list domain name + " " + scope
 */
    public void addDA(String url, String scope) { // add a DA to the DA list
        url   = url.trim();
        scope = scope.trim();
        if (url.equals("") || scope.equals("")) {
            append("Either URL or scope is empty!");
            return;
        }
        String item = new String(url+" "+scope);
        if (!daListModel.contains(item)) {
            daListModel.addElement(item);
            append("Adding <" + item + "> to DA list");
            daList.setSelectedIndex(0);
        }
    }

    public void valueChanged(ListSelectionEvent e) { // select DA
        if (e.getValueIsAdjusting()) return;
        if (e.getSource() == daList) {
            String tmp = (String)daList.getSelectedValue();
            if (tmp == null) return;
            int index = tmp.indexOf(" ");
            dirAgent = tmp.substring(0, index);
            append("Using " + dirAgent + " as DA");
            if (tcp != null) {  // close previous tcp connection 
                tcp.close();
                tcp = null;
            }
        }
    }

    private void daFail() { // DA failure
        append("Failure, no response from " + dirAgent);
        dirAgent = null;
        delDA();
    }

    private void delDA() { // delete a DA from the list
        int index = daList.getSelectedIndex();
        if (index != -1) {
            daListModel.remove(index);
            if (daListModel.size() > 0) {
                daList.setSelectedIndex(0);
            }
        }
    }

    public void clearDA() {
        daListModel.clear();
        dirAgent = null;
    }

    public void remoteDA(String domain) { // add a remote DA to the DA list
        String dummy, port="", host="";
        domain = domain.trim();
        if (domain.equals("")) {
            append("remote DNS domain is empty!");
            return;
        }
        try {
            DirContext ctx = new InitialDirContext();
            Attributes attrs = ctx.getAttributes("dns://cs.columbia.edu/_slpda._tcp."+domain, new String[] {"SRV"});
            NamingEnumeration attrlist = attrs.getAll();
            if (attrlist.hasMore()) {
                Attribute attr = (Attribute) attrlist.next();
                String val = (String) attr.get();
                StringTokenizer st = new StringTokenizer(val);
                if (st.hasMoreTokens()) dummy = st.nextToken();
                if (st.hasMoreTokens()) dummy = st.nextToken();
                if (st.hasMoreTokens()) port = st.nextToken();
                if (st.hasMoreTokens()) {
                    host = st.nextToken();
                    int index = host.length() - 1;
                    if (host.charAt(index)=='.') {
                        host = host.substring(0, index);
                    }
                }
            }
        } catch (Exception e) {
            System.err.println(e);
            return;
        }
        String item = new String(host+":"+port+" default");
        if (!daListModel.contains(item)) {
            daListModel.addElement(item);
            append("Adding <" + item + "> to DA list");
            daList.setSelectedIndex(0);
        }
    }

    class MyTableModel extends AbstractTableModel {
        final String[] columnNames = {"Extension Number",
                                      "Extension Name", 
                                      "Extension Content"};
        Object[][] data;
        int max_row_count = 10;
        int row_count = 0;

        MyTableModel () {
            data = new Object[max_row_count][3];
            for (int i=0; i<max_row_count; i++) {
                data[i][0] = new Integer(i+1);
                data[i][1] = " ";
                data[i][2] = " ";
            }
        }
        public int getColumnCount() {
            return columnNames.length;
        }
        
        public int getRowCount() {
            return row_count;
        }

        public int getMaxRowCount() {
            return max_row_count;
        }

        public void setRowCount(int n) {
            row_count = n;
            fireTableDataChanged();
        }

        public void clearRow(int i) {
            data[i][1] = " ";
            data[i][2] = " ";
        }

        public String getColumnName(int col) {
            return columnNames[col];
        }

        public Object getValueAt(int row, int col) {
            return data[row][col];
        }

        public boolean isCellEditable(int row, int col) {
            if (col >= 1) {
                return true;
            } else {
                return false;
            }
        }

        public Class getColumnClass(int c) {
            return getValueAt(0, c).getClass();
        }

        //-----------------------------------------------------------------
        // Don't implement this method unless your table's data can change.
        //-----------------------------------------------------------------
        public void setValueAt(Object value, int row, int col) {
            data[row][col] = value;
            fireTableCellUpdated(row, col);
        }
    }
}
